# coding=utf-8
"""
All commands that create & use the local copies of the Google Calendar data.
"""
import requests

from icalendar import Calendar

from ..config import config, ICAL_PATH
from ..utils import nice_format

from .parser import subparsers


def download(reset=None):
    if reset:
        config.pop('ical_url', None)

    try:
        ical_url = config['ical_url']
    except KeyError:
        print "To download the iCal file for analysis, you must give me the " \
              "public URL for it."
        print "Please go to the Google Calendar web interface " \
              ", 'Calendar Settings', and then copy the link address from " \
              "the ICAL button under 'Calendar Address'"
        ical_url = raw_input("Paste --> ")
        config['ical_url'] = ical_url

    print "Downloading private iCal file..."
    req = requests.get(ical_url, stream=True)

    if req.status_code != 200:
        print "Could not fetch iCal url - has it expired? "
        print "To change, run download --reset"
        print ical_url
        return False

    with open(ICAL_PATH, 'wb') as f:
        for chunk in req.iter_content():
            f.write(chunk)

    print "Download successful!"

    make_db()

    return True

download.parser = subparsers.add_parser(
    'download',
    description="Downloads the iCal that contains the whole of your Google "
                "Calendar, and then parses it into the local database"
)
download.parser.add_argument(
    '-r',
    '--reset',
    const=True,
    default=False,
    nargs='?',
    help="Pass this in to force re-pasting in the iCal url, if e.g. the url "
         " stored in lifelogger is no longer valid."
)
download.parser.set_defaults(func=download)


def make_db():
    from ..database import Event

    print "Converting iCal file into sqlite database..."

    with open(ICAL_PATH, 'rb') as f:
        ical_data = f.read()

    cal = Calendar.from_ical(ical_data)

    try:
        Event.drop_table()
    except:
        pass
    try:
        Event.create_table()
    except:
        pass

    for event in cal.walk("VEVENT"):
        Event.create_from_ical_event(event)

    print "Imported {} events.".format(
        Event.select().count()
    )

    return True

make_db.parser = subparsers.add_parser(
    'make_db',
    description="Parses the downloaded iCal file into the local sqlite "
                "database. Normally done when the download command is run, "
                "but may need re-running on changes to lifelogger."
)
make_db.parser.set_defaults(func=make_db)


def shell():
    from datetime import datetime, date
    from ..database import Event, regexp

    # This is just to avoid 'unused import' linter errors
    [datetime, date, Event, regexp]

    from IPython import embed
    embed()

shell.parser = subparsers.add_parser(
    'shell',
    description="Loads the local database and an IPython shell so you can "
                "manually search around the events using the 'peewee' ORM."
)
shell.parser.set_defaults(func=shell)


def sql(statement, separator):
    from ..database import conn
    statement = ' '.join(statement)

    cursor = conn.cursor()
    cursor.execute(statement)

    separator = {
        'comma': ',',
        'semicolon': ';',
        'tab': '\t',
    }[separator]

    # Header
    print separator.join([d[0] for d in cursor.description])

    # Data
    for row in cursor.fetchall():
        print separator.join([str(v) for v in row])

sql.parser = subparsers.add_parser(
    'sql',
    description="Execute a SQL statement direct on the db and output results "
                "as csv."
)
sql.parser.add_argument(
    'statement',
    nargs="+",
    type=unicode,
    help="The SQL statement."
)
sql.parser.add_argument(
    '-s',
    '--separator',
    nargs="?",
    type=unicode,
    default="comma",
    choices=['comma', 'semicolon', 'tab'],
    help="Separator for the output - default comma."
)
sql.parser.set_defaults(func=sql)


def list_command(filter_re):
    filter_re = ' '.join(filter_re)
    from ..database import Event, regexp

    events = Event.select().where(regexp(Event.summary, filter_re))

    for event in events:
        print event.display()

    return True

list_command.parser = subparsers.add_parser(
    'list',
    description="Lists the events that match a given regex."
)
list_command.parser.add_argument(
    'filter_re',
    nargs="+",
    type=unicode,
    help="The regex to filter events by."
)
list_command.parser.set_defaults(func=list_command)


def csv(filter_re, separator, varnames):
    filter_re = ' '.join(filter_re)

    varnames = varnames.split(',')

    separator = {
        'comma': ',',
        'semicolon': ';',
        'tab': '\t',
    }[separator]

    from ..database import Event, regexp

    events = Event.select().where(regexp(Event.summary, filter_re))

    # Header
    print separator.join(varnames)

    # Data
    for event in events:
        print separator.join([
            nice_format(event.get_var(varname)) for varname in varnames
        ])


csv.parser = subparsers.add_parser(
    'csv',
    description="Used to output properties of events that a given filter as "
                "CSV data."
)
csv.parser.add_argument(
    '-s',
    '--separator',
    nargs="?",
    type=unicode,
    default="comma",
    choices=['comma', 'semicolon', 'tab'],
    help="Separator for the output - default comma."
)
csv.parser.add_argument(
    '-v',
    '--varnames',
    nargs="?",
    type=unicode,
    default="start,end,summary",
    help="A comma-separated list of the Event variables to output (options: "
         "start, end, summary, duration_seconds, duration_minutes, "
         "duration_hours, units, percentage, kg, mg). "
         "Defaults to 'start,end,summary'."
)
csv.parser.add_argument(
    'filter_re',
    nargs="+",
    type=unicode,
    help="The regex to filter events by."
)
csv.parser.set_defaults(func=csv)



# coding=utf-8
import argparse

description = """
lifelogger: Track your life like a pro on Google Calendar via your terminal.

An all-in-one solution for adding events quickly, downloading them all into a
local SQLite database, and analyzing/exporting them.

To discover more about a command, run it with '--help', e.g.
'lifelogger download --help'.

Try starting with the 'add', 'download', and 'list' commands.
"""

# The main parser
parser = argparse.ArgumentParser(
    prog="lifelogger",
    description=description,
    epilog="Enjoy! Adam"
)

subparsers = parser.add_subparsers()



# coding=utf-8
"""
Handles connecting to Google API and authenticating.
"""
import argparse
import httplib2
import os
import sys

from apiclient import discovery as apc_discovery
from oauth2client import file as oa2c_file
from oauth2client import client as oa2c_client
from oauth2client import tools as oa2c_tools

CLIENT_SECRETS_PATH = os.path.join(
    os.path.abspath(os.path.dirname(__file__)),
    'client_secrets.json'
)

# Parser for command-line arguments - only used to get defaults
parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    parents=[oa2c_tools.argparser]
)


FLOW = oa2c_client.flow_from_clientsecrets(
    CLIENT_SECRETS_PATH,
    scope=[
        'https://www.googleapis.com/auth/calendar',
        'https://www.googleapis.com/auth/calendar.readonly',
    ],
    message="Error - client_secrets.json file missing."
)


def connect():
    from config import config, DATA_PATH

    flags = parser.parse_args([])

    # If the credentials don't exist or are invalid run through the native client
    # flow. The Storage object will ensure that if successful the good
    # credentials will get written back to the file.
    storage = oa2c_file.Storage(os.path.join(DATA_PATH, 'google_auth.json'))
    credentials = storage.get()
    if credentials is None or credentials.invalid:
        credentials = oa2c_tools.run_flow(FLOW, storage, flags)

    # Create an httplib2.Http object to handle our HTTP requests and authorize it
    # with our good Credentials.
    http = httplib2.Http()
    http = credentials.authorize(http)

    # Construct the service object for the interacting with the Calendar API.
    try:
        service = apc_discovery.build('calendar', 'v3', http=http)
    except httplib2.ServerNotFoundError:
        sys.stderr.write("Error: server not found - are you connected to the internet?\n")
        sys.exit()

    if 'calendar_id' not in config:
        all_cals = service.calendarList().list().execute()['items']
        primary_cal = [cal for cal in all_cals
                       if 'primary' in cal and cal['primary']][0]
        config['calendar_id'] = primary_cal['id']

    if 'timezone' not in config:
        settings = service.settings().list().execute()['items']
        settings = dict([(item['id'], item['value']) for item in settings])
        config['timezone'] = settings.get('timezone', "Europe/London")

    return service



# coding=utf-8
import re
from datetime import datetime, time

from peewee import CharField, DateTimeField, Expression, Model, SqliteDatabase

from .config import DB_PATH
from .utils import blue, highlight_tags, pink


# Add regex function to SqliteDatabase
OP_REGEXP = 'regexp'


def regexp(lhs, rhs):
    return Expression(lhs, 'regexp', rhs)

SqliteDatabase.register_ops({OP_REGEXP: 'REGEXP'})


# Create database reference
db = SqliteDatabase(DB_PATH)


# Define REGEXP function in sqlite database connection
conn = db.get_conn()


def regex_matches(regex, string):
    return bool(re.search(regex, string, flags=re.IGNORECASE))


conn.create_function('REGEXP', 2, regex_matches)


# Models
class Event(Model):
    summary = CharField()
    start = DateTimeField()
    end = DateTimeField()
    description = CharField()

    class Meta:
        database = db
        indexes = (
            (('summary'), False),
            (('start'), False),
            (('end'), False),
        )
        order_by = ('start',)

    @classmethod
    def create_from_ical_event(cls, ical_event):
        start = normalized(ical_event.get('dtstart').dt)
        end = ical_event.get('dtend')

        # 0-minute events have no end
        if end is not None:
            end = normalized(end.dt)
        else:
            end = start

        return cls.create(
            summary=ical_event.get('summary'),
            start=start,
            end=end,
            description=ical_event.get('description', ''),
        )

    def __unicode__(self):
        return u"{}    {}    {}".format(
            self.start,
            self.end,
            self.summary,
        )

    def display(self):
        # Pretty tabular formatting.
        start = self.start.strftime('%Y %b %d %H:%M')
        end = self.end.strftime('%b %d %H:%M')

        out = u"{}\t{}\t{}".format(
            blue(start),
            blue(end),
            highlight_tags(self.summary)
        )

        if self.description:
            out += u"\n\t" + pink(format(self.description)).replace("\n", "\n\t")

        return out

    @property
    def start_date(self):
        return self.start.date()

    @property
    def duration_seconds(self):
        return (self.end - self.start).total_seconds()

    @property
    def duration_minutes(self):
        return self.duration_seconds / 60.0

    @property
    def duration_hours(self):
        return self.duration_seconds / (3600.0)

    @property
    def duration_days(self):
        return self.duration_seconds / (3600.0 * 24)

    def get_var(self, varname):
        if hasattr(self, varname):
            return getattr(self, varname)
        else:
            return self.equality_property(varname)

    def equality_property(self, key):
        """
        Extracts a property from the event of the form key=value, e.g. units=3
        """
        try:
            return re.search('\\b%s=(\S+)\\b' % key, self.summary).group(1)
        except (AttributeError):
            raise ValueError("Event {} doesn't match for property {}".format(self, key))

    @property
    def percentage(self):
        # Used for #bodyfat measurements
        return self.measurement_property('percentage', '\\b([0-9.]+)%')

    @property
    def kg(self):
        # Used for #weight measurements
        return self.measurement_property('kg', '([0-9.]+)kg\\b')

    @property
    def mg(self):
        # Used for drug intake, e.g. #caffeine measurements
        return self.measurement_property('mg', '([0-9.]+)mg\\b')

    def measurement_property(self, units, regex):
        """
        Extracts a property from the event description of the form Xunits,
        e.g. 80kg
        """
        try:
            return float(
                re.search(regex, self.summary).group(1)
            )
        except (ValueError, AttributeError):
            raise ValueError("Event {} doesn't match for property {}".format(self, units))


def normalized(dt):
    # Fix the broken API for ical events - dt may be a date or datetime, so
    # make sure it is a datetime
    if not isinstance(dt, datetime):
        dt = datetime.combine(dt, time(0))

    if dt.tzinfo is not None:
        dt = dt.replace(tzinfo=None)

    return dt
